#include "Metrics.hpp"

// ESP32 specific
#include "freertos/FreeRTOS.h"
// #include "freertos/semphr.h"
#include "freertos/task.h"

// ESP32 specific
#include "nvs.h"
#include "nvs_flash.h"
#include "esp_err.h"
#include "esp_log.h"

// qthing stuff
#include <qthing>

// C++ system level
#include <climits>
#include <cinttypes>

// project specific
#include "../SpiderLib/SNTP.hpp"


using namespace qthing;
using json = nlohmann::json;


static const char* TAG = "Metrics";


static uint8_t nextTask = 0;
static void methodTaskWrapper(void* parameter) {
  std::function<void()>* f = (std::function<void()>*)parameter;
  (*f)();
}


namespace SiliconTorch {

  namespace Metrics {

    uint32_t uptime_ms() {
      return (uint32_t)(esp_timer_get_time() / 1000L);
    }

    Metrics::Metrics(const std::string& nameSpace) : nameSpace(nameSpace) {

      nvs_flash_init();
      nvs_open(nameSpace.c_str(), NVS_READWRITE, &this->nvs);

      nvs_get_u32(this->nvs, "bootCnt", &this->bootCnt);
      bootCnt++;

      nvs_get_u64(this->nvs, "lifetime", &this->lifeTime);
      lifeTimeSaved = uptime_ms();


      nvs_get_u64(this->nvs, "birth", &this->birth);
      lifeTimeSaved = uptime_ms();

      if (this->birth > 0) {
        ESP_LOGI(TAG, "Found my birth timestamp[ %lld ]", this->birth);

        SpiderLib::SNTP::instance().registerTimeSyncHandler([&](){
          uint64_t now = (uint64_t)std::time(NULL);

          float age = (float)(now - this->birth);
          age /= 24.0f * 60.0f * 60.0f;

          ESP_LOGI(TAG, "I'm at least days[ %.2f ] old", age);
        });
      } else {
        SpiderLib::SNTP::instance().registerTimeSyncHandler([&](){

          this->birth = (uint64_t)std::time(NULL);

          esp_err_t err = nvs_set_u64(this->nvs, "birth", this->birth);

          if (err == ESP_OK) ESP_LOGI(TAG, "Just carved my birth timestamp[ %lld ] 🎂 into silicon", this->birth);
          else ESP_LOGE(TAG, "Couldn't save birth time: error[ %d ]", err);
        });
      }


      char taskName[64];
      snprintf(taskName, sizeof(taskName), "Metrics[%i]", nextTask++);  // TODO: increment thread-safe!

      std::function<void()> f = std::bind(&SiliconTorch::Metrics::Metrics::tickTask, this);
      std::function<void()>* _f = new std::function<void()>(f);
      xTaskCreatePinnedToCore(methodTaskWrapper, taskName, 8192, (void*)_f, 1, NULL, 0);


      snprintf(taskName, sizeof(taskName), "bootCnt[%i]", nextTask++);  // TODO: increment thread-safe!

      std::function<void()> f2 = std::bind(&SiliconTorch::Metrics::Metrics::saveBootCntTask, this);
      std::function<void()>* _f2 = new std::function<void()>(f2);
      xTaskCreatePinnedToCore(methodTaskWrapper, taskName, 8192, (void*)_f2, 1, NULL, 0);


      std::function<void(const std::string&)> getMetrics = [&](const std::string& ignored) {
        publishData();
      };

      add_message_callback(genDeviceTopic("listener/get"), getMetrics);
      add_message_callback(std::string("service/") + nameSpace + std::string("/listener/get"), getMetrics);
      add_mqtt_connected_callback(std::bind(&SiliconTorch::Metrics::Metrics::publishData, this));
    }


    void Metrics::registerMetric(const std::string& shortName, const std::string& fullName) {

      auto result = counters.find(shortName);
      if (result != counters.end()) {
        ESP_LOGW(TAG, "Can't register Metric[ %s ]: Already exists", shortName.c_str());
      } else {
        CounterMetric* counter = new CounterMetric(shortName, fullName);

        counter->load(&this->nvs);
        counters[shortName] = counter;
      }
    }

    void Metrics::publishData() {
      json j;

      for (auto it = counters.begin(); it != counters.end(); ++it) {
        CounterMetric* metric = it->second;
        j[metric->fullName.c_str()] = metric->getValue();
      }

      j["reset"]    = bootCnt;
      j["uptime"]   = uptime_ms() / 1000UL;
      j["lifetime"] = (lifeTime + (uptime_ms() - lifeTimeSaved)) / 1000UL;

      if (birth > 0) j["birth"] = birth;

      publish_message(genDeviceTopic("metrics"), j.dump().c_str());
      lastPublished = uptime_ms();
    }

    void Metrics::saveData() {

      for (auto it = counters.begin(); it != counters.end(); ++it)
        it->second->save(&this->nvs);


      uint32_t now = uptime_ms();

      lifeTime += now - lifeTimeSaved;
      lifeTimeSaved = now;

      esp_err_t err = nvs_set_u64(this->nvs, "lifetime", this->lifeTime);

      if (err != ESP_OK) {
        const char* errorName = esp_err_to_name(err);
        ESP_LOGE(TAG, "Error[ %s ] saving LifeTimeMetric[ %lli ]", errorName, lifeTime);
      }

      nvs_commit(this->nvs);

      lastSaved = uptime_ms();
    }

    void Metrics::tickTask() {

      TickType_t lastWakeTime = xTaskGetTickCount();

      while (true) {

        uint32_t now = uptime_ms();

        if ((now - lastSaved    ) / 1000 > saveInterval   ) saveData();
        if ((now - lastPublished) / 1000 > publishInterval) publishData();

        vTaskDelayUntil(&lastWakeTime, tickPeriod / portTICK_PERIOD_MS);
      }
    }

    void Metrics::saveBootCntTask() {

      vTaskDelay(bootCntSaveUptime * 1000 / portTICK_PERIOD_MS);

      ESP_LOGI(TAG, "Saving bootCnt[ %i ] at uptime[ %i ]", bootCnt, uptime_ms() / 1000);

      nvs_set_u32(this->nvs, "bootCnt", this->bootCnt);
      nvs_commit(this->nvs);

      vTaskDelete(NULL);
    }


    uint64_t Metrics::getCounterValue(const std::string& shortName) {

      auto result = counters.find(shortName);

      if (result != counters.end()) {
        return result->second->getValue();
      } else if (shortName.compare("uptime") == 0) {
        return uptime_ms() / 1000UL;
      } else if (shortName.compare("lifetime") == 0) {
        return (lifeTime + (uptime_ms() - lifeTimeSaved)) / 1000UL;
      } else if (shortName.compare("bootCnt") == 0) {
        return bootCnt;
      } else {
        return 0;  // TODO: log warning…?
      }
    }

    std::string Metrics::genDeviceTopic(const char* suffix) {
      return std::string(DEVICE_NAMESPACE) + nameSpace + std::string("/") + std::string(suffix);
    }

    Incrementer Metrics::generateMetricIncrementer(const std::string& shortName) {

      auto result = counters.find(shortName);

      if (result != counters.end()) {
        CounterMetric* counter = result->second;

        return [counter]() { counter->inc(); };
      } else {
        return [](){};  // TODO: log warning…?
      }
    }


    CounterMetric::CounterMetric(const std::string& shortName, const std::string& fullName) : shortName(shortName), fullName(fullName) {}

    void CounterMetric::load(nvs_handle_t* nvs) {
      nvs_get_u64(*nvs, shortName.c_str(), &this->value);
    }

    void CounterMetric::save(nvs_handle_t* nvs) {
      if (dirty) forceSave(nvs);
    }

    void CounterMetric::forceSave(nvs_handle_t* nvs) {
      dirty = false;

      esp_err_t err = nvs_set_u64(*nvs, shortName.c_str(), this->value);

      if (err != ESP_OK) {
        const char* errorName = esp_err_to_name(err);
        ESP_LOGE(TAG, "Error[ %s ] saving Metric[ %s ]", errorName, fullName.c_str());
      }
    }

    uint64_t CounterMetric::getValue() {
      return value;
    }


    // TODO: synchronize!
    void CounterMetric::inc() {
      if (value < ULLONG_MAX) {
        dirty = true;
        value++;
      }
    }
    
    // TODO: synchronize!
    void CounterMetric::dec() {
      if (value > 0) {
        dirty = true;
        value--;
      }
    }

    // pre-increment
    uint64_t& CounterMetric::CounterMetric::operator++() {
      inc(); return value;
    }

    // pre-decrement
    uint64_t& CounterMetric::CounterMetric::operator--() {
      dec(); return value;
    }

    // post-increment
    uint64_t CounterMetric::CounterMetric::operator++(int ignored) {
      uint64_t out = value; inc(); return out;
    }

    // post-decrement
    uint64_t CounterMetric::CounterMetric::operator--(int ignored) {
      uint64_t out = value; dec(); return out;
    }
  }
}
